#ifndef __GRAPH
#define __GRAPH

#ifndef NOSTUB
#ifndef DOORS
#ifndef USE_KERNEL
#include <nostub.h>
#else
#include <doors.h>
#endif
#endif
#endif

#ifndef __GR_TYPES
#define __GR_TYPES

enum Attrs {A_REVERSE,A_NORMAL,A_XOR,A_SHADED,A_REPLACE,A_OR,A_AND,A_THICK1,
            A_SHADE_V,A_SHADE_H,A_SHADE_NS,A_SHADE_PS};

enum Fonts {F_4x6,F_6x8,F_8x10};

enum BoxAttrs {B_NORMAL=0x10,B_ROUNDED=0x20,B_DOUBLE=0x40,B_CUT=0x80};

typedef unsigned char SCR_COORDS;
typedef short WIN_COORDS;

#ifndef __SCR_RECT
#define __SCR_RECT

typedef union
  {
    struct
      {
        unsigned char x0,y0,x1,y1;
      } xy;
    unsigned long l;
  } SCR_RECT;

#endif

#ifndef __WIN_RECT
#define __WIN_RECT

typedef struct
  {
    short x0,y0,x1,y1;
  } WIN_RECT;

#endif

#ifndef __ICON
#define __ICON

typedef struct {unsigned short i[16];} ICON;
typedef unsigned short *pICON;

#endif

#define BITMAP_HDR_SIZE   4

typedef struct
  {
    unsigned short NumRows,NumCols;
    unsigned char Data[0];               /* GNU C extension for [] */
  } BITMAP;

#endif

typedef struct
  {
    void *ScrAddr;
    unsigned char XMax,YMax;
    short CurFont,CurAttr,CurX,CurY;
    SCR_RECT CurClip;
  } SCR_STATE;

typedef struct
  {
    unsigned char Count;
    struct
      {
        unsigned char Attr,x0,y0,x1,y1;
      } Data[0];                         /* GNU C extension for [] */
  } MULTI_LINE;

#ifndef __BOOL
#define __BOOL

enum Bool {FALSE,TRUE};

#endif

#define LCD_MEM ((void*)0x4C00)
#define LCD_SIZE 3840

typedef char LCD_BUFFER[LCD_SIZE];

#define ClearScreen _rom_call(void,(void),19E)
#define ClrScr _rom_call(void,(void),19E)
#define GetPix _rom_call(short,(short,short),19F)
#define DrawStr _rom_call(void,(short,short,char*,short),1A9)
#define DrawStrXY _rom_call(void,(short,short,char*,short),1A9)
#define DrawChar _rom_call(void,(short,short,char,short),1A4)
#define DrawLine _rom_call(void,(short,short,short,short,short),1A7)
#define DrawPix _rom_call(void,(short,short,short),1A8)
#define DrawFkey _rom_call(void,(short,short,short,short),1A5)
#define DrawIcon _rom_call(void,(short,short,void*,short),1A6)
#define FontGetSys _rom_call(unsigned char,(void),18E)
#define FontSetSys _rom_call(unsigned char,(short),18F)
#define FontCharWidth _rom_call(short,(short),190)
#define DrawStrWidth _rom_call(short,(char*,short),197)
#define SetCurAttr _rom_call(short,(short),19A)
#define SetCurClip _rom_call(void,(SCR_RECT*),19B)
#define MoveTo _rom_call(void,(short,short),19D)
#define LineTo _rom_call(void,(short,short),19C)
#define DrawTo _rom_call(void,(short,short),19C)
#define DrawClipPix _rom_call(void,(short,short),194)
#define DrawClipChar _rom_call(void,(short,short,short,SCR_RECT*,short),191)
#define DrawClipLine _rom_call(void,(WIN_RECT*,SCR_RECT*,short),193)
#define DrawClipRect _rom_call(void,(WIN_RECT*,SCR_RECT*,short),195)
#define DrawClipEllipse _rom_call(void,(short,short,short,short,SCR_RECT*,short),192)
#define DrawMultiLines _rom_call(void,(short,short,void*),196)
#define FillTriangle \
  _rom_call(void,(short,short,short,short,short,short,SCR_RECT*,short),198)
#define FillLines2 _rom_call(void,(WIN_RECT*,WIN_RECT*,SCR_RECT*,short),199)
#define BitmapGet _rom_call(void,(SCR_RECT*,void*),185)
#define BitmapInit _rom_call(void,(SCR_RECT*,void*),186)
#define BitmapPut _rom_call(void,(short,short,void*,SCR_RECT*,short),187)
#define BitmapSize _rom_call(unsigned short,(SCR_RECT*),188)
#define ScrRectFill _rom_call(void,(SCR_RECT*,SCR_RECT*,short),189)
#define ScrRectScroll _rom_call(void,(SCR_RECT*,SCR_RECT*,short,short),18B)
#define ScrRectShift _rom_call(void,(SCR_RECT*,SCR_RECT*,short,short),18C)
#define ScrRectOverlap _rom_call(short,(SCR_RECT*,SCR_RECT*,SCR_RECT*),18A)
#define QScrRectOverlap _rom_call(short,(SCR_RECT*,SCR_RECT*),18D)
#define ScrToWin _rom_call(WIN_RECT*,(SCR_RECT*),2D)
#define ScrToHome _rom_call(SCR_RECT*,(SCR_RECT*),2E)
#define PortSet _rom_call(void,(void*,short,short),1A2)
#define PortRestore _rom_call(void,(void),1A3)
#define SaveScrState _rom_call(void,(void*),1A0)
#define RestoreScrState _rom_call(void,(void*),1A1)

#ifdef DOORS

void ClearScreen(void);
void ClrScr(void);
short GetPix(short,short);
void DrawStr(short,short,char*,short);
void DrawStrXY(short,short,char*,short);
void DrawChar(short,short,char,short);
void DrawLine(short,short,short,short,short);
void DrawPix(short,short,short);
void DrawFkey(short,short,short,short);
void DrawIcon(short,short,void*,short);
unsigned char FontGetSys(void);
unsigned char FontSetSys(short);
short FontCharWidth(short);
short DrawStrWidth(char*,short);
short SetCurAttr(short);
void SetCurClip(SCR_RECT*);
void MoveTo(short,short);
void LineTo(short,short);
void DrawTo(short,short);
void DrawClipPix(short,short);
void DrawClipChar(short,short,short,SCR_RECT*,short);
void DrawClipLine(WIN_RECT*,SCR_RECT*,short);
void DrawClipRect(WIN_RECT*,SCR_RECT*,short);
void DrawClipEllipse(short,short,short,short,SCR_RECT*,short);
void DrawMultiLines(short,short,void*); 
void FillTriangle(short,short,short,short,short,short,SCR_RECT*,short);
void FillLines2(WIN_RECT*,WIN_RECT*,SCR_RECT*,short);
void BitmapGet(SCR_RECT*,void*);
void BitmapInit(SCR_RECT*,void*);
void BitmapPut(short,short,void*,SCR_RECT*,short);
unsigned short BitmapSize(SCR_RECT*);
void ScrRectFill(SCR_RECT*,SCR_RECT*,short);
void ScrRectScroll(SCR_RECT*,SCR_RECT*,short,short);
void ScrRectShift(SCR_RECT*,SCR_RECT*,short,short);
short ScrRectOverlap(SCR_RECT*,SCR_RECT*,SCR_RECT*);
short QScrRectOverlap(SCR_RECT*,SCR_RECT*);
WIN_RECT *ScrToWin(SCR_RECT*);
SCR_RECT *ScrToHome(SCR_RECT*);
void PortSet(void*,short,short);
void PortRestore(void);
void SaveScrState(void*);
void RestoreScrState(void*);

#endif

#ifndef __MAKEWINRECT
#define __MAKEWINRECT

#define MakeWinRect _rom_call(WIN_RECT*,(short,short,short,short),2C)

#ifdef DOORS

WIN_RECT *MakeWinRect(short,short,short,short);

#endif

#endif

#ifdef NOSTUB

#define ScrRect ((SCR_RECT*)(*((long*)(*((long*)0xC8)+0x2F*4))))

#else

#define ScrRect (&_ROM_CALL_2F)
extern SCR_RECT _ROM_CALL_2F;

void *_ROM_CALL_26A(void*,void*,long);

#endif

#define LCD_save(b) \
  ((void)(_rom_call(void,(),26A)((void*)(b),LCD_MEM,(long)LCD_SIZE)))
#define LCD_restore(b) \
  ((void)(_rom_call(void,(),26A)(LCD_MEM,(void*)(b),(long)LCD_SIZE)))

#endif
