#ifndef __TIMATH
#define __TIMATH

#ifndef NOSTUB
#ifndef DOORS
#ifndef USE_KERNEL
#include <nostub.h>
#else
#include <doors.h>
#endif
#endif
#endif

#ifndef __BOOL
#define __BOOL

enum Bool {FALSE,TRUE};

#endif

#ifndef __BCD
#define __BCD

typedef float ti_float;

typedef struct
  {
    unsigned short exponent;
    unsigned long long mantissa;
  } bcd;

#endif

#ifndef __SABS
#define __SABS

#define abs(x) ({typeof(x) __x=(x); __x>=0?__x:-__x;})

#endif

long labs(long);    /* This function is open-coded in GCC itself */

#define FLT(x,y...) ((sizeof(#y)==1)?x##.0:x##.##y)
#define FLT_NEG(x,y...) ((sizeof(#y)==1)?-x##.0:-x##.##y)

#define FEXP(x,y) (*(float*)&(bcd){0x4000+y,0x##x##LL<<4*(17-sizeof(#x))})
#define FEXP_NEG(x,y) (*(float*)&(bcd){0xC000+y,0x##x##LL<<4*(17-sizeof(#x))})

#ifndef __FP_SPECIALS
#define __FP_SPECIALS

#define ZERO (0.0)
#define UNSIGNED_ZERO (0.0)
#define POSITIVE_ZERO (*(float*)&(bcd){0,0})
#define NEGATIVE_ZERO (*(float*)&(bcd){0x8000,0})

#define UNSIGNED_INF (*(float*)&(bcd){0x7FFF,0xAA00CC0000000000})
#define POSITIVE_INF (*(float*)&(bcd){0x7FFF,0xAA00BB0000000000})
#define NEGATIVE_INF (*(float*)&(bcd){0xFFFF,0xAA00BB0000000000})

#define NAN (*(float*)&(bcd){0x7FFF,0xAA00000000000000})

#define ldexp10(x,e) ({float __f=x;((bcd*)&__f)->exponent+=e;__f;})

#endif

#define ONE (1.0)
#define TWO (2.0)
#define THREE (3.0)
#define FOUR (4.0)
#define FIVE (5.0)
#define TEN (10.0)
#define HALF (0.5)
#define MINUS_ONE (-1.0)

#ifndef __PI
#define __PI

#define PI (3.141592653589793)
#define HALF_PI (1.570796326794897)

#endif

#define fcmp _rom_call(long,(float,float),BB)
#define trunc _rom_call(long,(float),BC)

#define sincos _rom_call(void,(float,short,float*,float*),286)
#define itrig _rom_call(void,(short,short,float*,float*),28A)
#define trig _rom_call(void,(short,short,float*,float*,float*,float*),28B)

#ifndef __CFNCS
#define __CFNCS

#define cacos _rom_call(void,(float,float,float*,float*),13A)
#define casin _rom_call(void,(float,float,float*,float*),13B)
#define catan _rom_call(void,(float,float,float*,float*),13C)
#define cacosh _rom_call(void,(float,float,float*,float*),13D)
#define casinh _rom_call(void,(float,float,float*,float*),13E)
#define catanh _rom_call(void,(float,float,float*,float*),13F)
#define ccos _rom_call(void,(float,float,float*,float*),140)
#define csin _rom_call(void,(float,float,float*,float*),141)
#define ctan _rom_call(void,(float,float,float*,float*),142)
#define ccosh _rom_call(void,(float,float,float*,float*),143)
#define csinh _rom_call(void,(float,float,float*,float*),144)
#define ctanh _rom_call(void,(float,float,float*,float*),145)
#define csqrt _rom_call(void,(float,float,float*,float*),146)
#define cln _rom_call(void,(float,float,float*,float*),147)
#define clog10 _rom_call(void,(float,float,float*,float*),148)
#define cexp _rom_call(void,(float,float,float*,float*),149)

#endif

#define is_float_infinity _rom_call(short,(float),2FF)
#define is_float_negative_zero _rom_call(short,(float),300)
#define is_float_positive_zero _rom_call(short,(float),301)
#define is_float_signed_infinity _rom_call(short,(float),302)
#define is_float_transfinite _rom_call(short,(float),303)
#define is_float_unsigned_inf_or_nan _rom_call(short,(float),304)
#define is_float_unsigned_zero _rom_call(short,(float),305)

#ifndef __FQUERY
#define __FQUERY

#define is_nan _rom_call(short,(float),306)

#define is_inf _rom_call(short,(float),2FF)
#define is_nzero _rom_call(short,(float),300)
#define is_pzero _rom_call(short,(float),301)
#define is_sinf _rom_call(short,(float),302)
#define is_transfinite _rom_call(short,(float),303)
#define is_uinf_or_nan _rom_call(short,(float),304)
#define is_uzero _rom_call(short,(float),305)

#endif

#define float_class _rom_call(short,(float),2FA)

#define fpisanint _rom_call(short,(long long*,short),172)
#define fpisodd _rom_call(short,(long long*,short),173)

#ifdef DOORS

long fcmp(float,float);
long trunc(float);

void sincos(float,short,float*,float*);
void itrig(short,short,float*,float*);
void trig(short,short,float*,float*,float*,float*);

void cacos(float,float,float*,float*);
void casin(float,float,float*,float*);
void catan(float,float,float*,float*);
void cacosh(float,float,float*,float*);
void casinh(float,float,float*,float*);
void catanh(float,float,float*,float*);
void ccos(float,float,float*,float*);
void csin(float,float,float*,float*);
void ctan(float,float,float*,float*);
void ccosh(float,float,float*,float*);
void csinh(float,float,float*,float*);
void ctanh(float,float,float*,float*);
void csqrt(float,float,float*,float*);
void cln(float,float,float*,float*);
void clog10(float,float,float*,float*);
void cexp(float,float,float*,float*);

short is_float_infinity(float);
short is_float_negative_zero(float);
short is_float_positive_zero(float);
short is_float_signed_infinity(float);
short is_float_transfinite(float);
short is_float_unsigned_inf_or_nan(float);
short is_float_unsigned_zero(float);
short is_nan(float);

short float_class(float);

short fpisanint(long long*,short);
short fpisodd(long long*,short);

#endif

#ifndef __ATOF
#define __ATOF

#define atof ((float(*)(char*))(short[]){0x4E56,0xFFB0,0x48E7,0x1F3C,0x2878, \
  0xC8,0x266C,0x424,0x2813,0x486E,0xFFB0,0x206C,0x550,0x4E90,0x4A40,0x670C, \
  0x203C,0x7FFF,0xAA00,0x4281,0x4242,0x603C,0x2EAE,8,0x206C,0xF28,0x4E90, \
  0x2053,0x1A10,0xC05,0x7A,0x6602,0x5348,0x2E88,0x206C,0xBE4,0x4E90,0x2684, \
  0x206C,0x554,0x4E90,0x202E,0xFFF6,0x222E,0xFFFA,0x342E,0xFFFE,0xC05,0x7A, \
  0x6604,0x8C0,0x1F,0x4CEE,0x3CF8,0xFF8C,0x4E5E,0x4E75})

#endif

/* Now follows a set of interface macros for adapting GCC convention of */
/* returning floating point values to the convention expected by TIOS   */

#ifndef __FLOAT_INTERFACE
#define __FLOAT_INTERFACE

#define __BC ((float(*)())(short[]){0x4E56,-10,0x2F2E,26,0x2F2E,22,0x2F2E, \
  18,0x2F2E,14,0x2F2E,10,0x302E,8,0x2078,200,0x2070,0,0x4E90,0x202E,-10, \
  0x222E,-6,0x342E,-2,0x4E5E,0x4E75})

#define _tios_float_1(f,x,t) ((float(*)(short,t))__BC)(4*0x##f,x)
#define _tios_float_2(f,x,y,t1,t2) ((float(*)(short,t1,t2))__BC)(4*0x##f,x,y)

#endif

/* Functions listed below are implemented via interface macros */

#define fadd(x,y) _tios_float_2(B6,x,y,float,float)
#define fsub(x,y) _tios_float_2(B7,x,y,float,float)
#define fmul(x,y) _tios_float_2(B8,x,y,float,float)
#define fdiv(x,y) _tios_float_2(B9,x,y,float,float)
#define fneg(x) _tios_float_1(BA,x,float)
#define flt(x) _tios_float_1(BD,x,long)

#ifndef __FLTFNCS
#define __FLTFNCS

#define acos(x) _tios_float_1(F5,x,float)
#define asin(x) _tios_float_1(F6,x,float)
#define atan(x) _tios_float_1(F7,x,float)
#define atan2(x,y) _tios_float_2(F8,x,y,float,float)
#define cos(x) _tios_float_1(F9,x,float)
#define sin(x) _tios_float_1(FA,x,float)
#define tan(x) _tios_float_1(FB,x,float)
#define cosh(x) _tios_float_1(FC,x,float)
#define sinh(x) _tios_float_1(FD,x,float)
#define tanh(x) _tios_float_1(FE,x,float)
#define exp(x) _tios_float_1(FF,x,float)
#define log(x) _tios_float_1(100,x,float)
#define log10(x) _tios_float_1(101,x,float)
#define modf(x,y) _tios_float_2(102,x,y,float,float*)
#define pow(x,y) _tios_float_2(103,x,y,float,float)
#define sqrt(x) _tios_float_1(104,x,float)
#define ceil(x) _tios_float_1(105,x,float)

#define floor(x) _tios_float_1(107,x,float)
#define fmod(x,y) _tios_float_2(108,x,y,float,float)

#define asinh(x) _tios_float_1(287,x,float)
#define acosh(x) _tios_float_1(288,x,float)
#define atanh(x) _tios_float_1(289,x,float)

#define frexp10(x,y) _tios_float_2(2FB,x,y,float,short*)

#endif

#ifndef __FABS
#define __FABS

#define fabs(x) _tios_float_1(106,x,float)

#endif

#define round12(x) _tios_float_1(174,x,float)
#define round14(x) _tios_float_1(175,x,float)
#define round12_err(x,y) _tios_float_2(227,x,y,float,short)

#ifndef __HYPOT
#define __HYPOT

#define hypot(x,y) ({float __x=(x),__y=(y); \
  sqrt(fadd(fmul((__x),(__x)),fmul((__y),(__y))));})

#endif

#define float_to_bcd(a) ({float __a=(a);*(bcd*)&__a;})
#define bcd_to_float(a) ({bcd __a=(a);*(float*)&__a;})

#define bcd_var(a) (*(bcd*)&(a))

#define bcdadd(x,y) ({bcd __x=(x),__y=(y); \
  float __z=fadd(*(float*)&__x,*(float*)&__y);*(bcd*)&__z;})
#define bcdsub(x,y) ({bcd __x=(x),__y=(y); \
  float __z=fsub(*(float*)&__x,*(float*)&__y);*(bcd*)&__z;})
#define bcdmul(x,y) ({bcd __x=(x),__y=(y); \
  float __z=fmul(*(float*)&__x,*(float*)&__y);*(bcd*)&__z;})
#define bcddiv(x,y) ({bcd __x=(x),__y=(y); \
  float __z=fdiv(*(float*)&__x,*(float*)&__y);*(bcd*)&__z;})

#define bcdneg(x) ({bcd __x=(x);float __y=fneg(*(float*)&__x);*(bcd*)&__y;})
#define bcdbcd(x) ({float __x=flt(x);*(bcd*)&__x;})

#define bcdcmp(x,y) ({bcd __x=(x),__y=(y);fcmp(*(float*)&__x,*(float*)&__y);})
#define bcdlong(x) ({bcd __x=(x);trunc(*(float*)&__x);})

#define init_float() ((void)0)    /* for compatibility */

#endif

